"""Utility functions for aiding plotting."""

import logging

from bokeh import palettes
from bokeh.models import Label
from bokeh.plotting import figure
import numpy as np
import pandas as pd
import pkg_resources
from scipy import stats as sp_stats


def get_named_logger(name):
    """Create a logger with a name.

    :param name: name of logger.
    """
    name = name.ljust(10)[:10]  # so logging is aligned
    logger = logging.getLogger('{}.{}'.format(__package__, name))
    logger.name = name
    return logger


def set_basic_logging(level=logging.INFO):
    """Set basic log formatting.

    :returns: logger instance.
    """
    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=level)
    logger = logging.getLogger(__package__)
    logger.setLevel(level)
    return logger


def read_files(summaries):
    """Read a set of files and join to single dataframe."""
    dfs = list()
    if not isinstance(summaries, (list, tuple)):
        summaries = [summaries]
    for fname in sorted(summaries):
        dfs.append(pd.read_csv(fname, sep="\t"))
    return pd.concat(dfs)


class _colors:
    """Some colours that someone thought were nice."""

    cerulean = "#0084A9"
    not_black = "#001A21"
    feldgrau = "#455556"
    dim_gray = "#666666"
    light_cornflower_blue = "#90C5E7"
    dark_gray = "#B5AEA7"
    isabelline = "#F0EFED"
    medium_spring_bud = "#B8E986"
    cinnabar = "#EF4134"
    sandstorm = "#F5CC49"
    fandango = "#A53F96"
    green = "#17BB75"
    verdigris = "#54B8B1"


Colors = _colors()


class _ontColors:
    """ONT Brand colors."""

    BRAND_BLUE = '#0084A9'
    BRAND_BLUE_LIGHT = '#0084a91a'
    BRAND_LIGHT_BLUE = '#90C6E7'
    BRAND_BLACK = '#333333'
    BRAND_LIGHT_GREY = '#B5AEA7'
    BRAND_GREY = '#666666'
    BRAND_GREY_LIGHT = '#6666661f'

    BRAND_LOGO = pkg_resources.resource_filename(
        __package__, 'data/ont_logo.txt')
    with open(BRAND_LOGO, 'r', encoding="UTF-8") as fh:
        BRAND_LOGO = fh.read()


ont_colors = _ontColors()


class _ondColors:
    """OND Brand colors."""

    BRAND_BLUE = '#003E71'
    BRAND_BLUE_LIGHT = '#003e7136'
    BRAND_LIGHT_BLUE = '#90C6E7'
    BRAND_BLACK = '#333333'
    BRAND_LIGHT_GREY = '#B5AEA7'
    BRAND_GREY = '#666666'
    BRAND_GREY_LIGHT = '#6666661f'
    BRAND_RED = '#f45b69'

    BRAND_LOGO = pkg_resources.resource_filename(
        __package__, 'data/ond_logo.txt')
    with open(BRAND_LOGO, 'r', encoding="UTF-8") as fh:
        BRAND_LOGO = fh.read()


ond_colors = _ondColors()


class Limiter(object):
    """Helper class to determine limits of data."""

    def __init__(self, limits=None):
        """Initialize the limit gathering."""
        if limits is not None:
            self.min, self.max = limits
        else:
            self.min, self.max = +float('inf'), -float('inf')

    def accumulate(self, data):
        """Analyse data to update limits.

        :param data: new data to analyse.

        """
        if len(data) > 0:
            self.min = min(self.min, min(data))
            self.max = max(self.max, max(data))
        return self

    def fix(self, lower=None, upper=None):
        """Fix limits to new values.

        :param lower: new lower limit.
        :param upper: new upper limit.

        """
        if lower is not None:
            self.min = lower
        if upper is not None:
            self.max = upper
        return self

    def __repr__(self):
        """Return string representation of self."""
        return repr((self.min, self.max))


def pad(data):
    """Attempt to find a sensible plot bounds for a vector."""
    uniq = sorted(np.unique(data))
    pad = 0.5 * min(np.ediff1d(uniq))
    return uniq[0] - pad, uniq[-1] + pad


def kernel_density_estimate(x, step=0.2):
    """Kernel density to approximate distribution.

    :param x: data of which to find mode.
    :param step: discretization of KDE PDF.
    """
    # estimate bandwidth of kde, R's nrd0 rule-of-thumb
    hi = np.std(x, ddof=1)
    q75, q25 = np.percentile(x, [75, 25])
    iqr = q75 - q25
    lo = min(hi, iqr/1.34)
    if not ((lo == hi) or (lo == abs(x[0])) or (lo == 1)):
        lo = 1
    bw = 0.9 * lo * len(x)**-0.2

    # create a KDE
    x_grid = np.arange(min(x), max(x), step)
    kernel = sp_stats.gaussian_kde(x, bw_method=bw)
    pdf = kernel(x_grid)
    return x_grid, pdf


def choose_palette(ncolours):
    """Choose colour palette.

    :param: ncolours: number of colours.
    """
    if ncolours <= 20:
        # Category20 contains dark/light pairs, avoid using these
        pal = palettes.Category20[20]
        palette = pal[0::2] + pal[1::2]
        cols = palette[0:ncolours]
    elif ncolours <= 256:
        cols = palettes.Viridis256[::256//ncolours]
    else:
        raise ValueError(
            "Cannot create colour palette with more than 256 colours.")
    return cols


def emptyPlot(**kwargs):
    """Empty plot for when all else fails.

    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.
    """
    p = figure(
        plot_width=300,
        plot_height=300,
        **kwargs
    )
    failLabel = Label(
        x=125, y=125, x_units='screen',
        y_units='screen',
        text=str('Failed to plot'),
        text_font_style='italic',
        text_align='center')

    p.circle([1, 2], [3, 4], fill_color='white', line_color='white')
    p.add_layout(failLabel)
    return p


def plot_wrapper(func):
    """Test for function and output empty graph if fails.

    :param func:  name of plotting function from aplanat.
    :param args: the arguments provided.
    :param kwargs: other optional key word arguments.

    :returns: plot.

    """
    def wrapper_accepting_arguments(*args, **kwargs):
        """Argument wrapper for decorator."""
        logger = get_named_logger('PlotWrap')
        try:
            p = (func(*args, **kwargs))
            return p
        except Exception as e:

            try:
                p = emptyPlot(**kwargs)
                logger.exception('Plotting user plot failed with: '+str(e))
                return p

            except Exception as f:
                p = emptyPlot()
                logger.exception(
                    'Plotting templated Empty plot failed with: ' + str(f))
                return p

    return wrapper_accepting_arguments



"""Parse single and multiple outputs from bcftools stats into dataframes."""

from collections import defaultdict
import os
import re

import pandas as pd


def split_blocks(fname):
    """Split lines of a file into blocks based on comment lines."""
    comment = list()
    data = list()
    state = None
    with open(fname, 'r') as fh:
        # TODO: handle empty tables
        for line in fh.readlines():
            if line.startswith('#'):
                if state != 'comment' and state is not None:
                    yield comment, data
                    comment, data = list(), list()
                state = 'comment'
                comment.append(line.strip('# ').rstrip())
            else:
                state = 'data'
                data.append(line.rstrip())
        yield comment, data


def parse_bcftools_stats(fname):
    """Parse `bcftools stats` output.

    :param fname: file to parse.
    """
    tables = dict()
    for comment, data in split_blocks(fname):
        fields = [x.rstrip() for x in re.split(r'\[\d+\]', comment[-1])]
        section, fields = fields[0], fields[1:]
        rows = list()
        for d in data:
            items = d.split('\t')
            if items[0] != section:
                raise ValueError("first data field not equal to section key")
            rows.append(items[1:])
        tables[section] = pd.DataFrame(rows, columns=fields)
    # now some special handling
    SN = tables['SN']
    SN['key'] = SN['key'].apply(
        lambda x: x.replace('number of ', '').rstrip(':'))
    tables['SN'] = pd.DataFrame(
        SN.pivot(index='id', columns='key', values='value').to_records())
    return tables


def parse_bcftools_stats_multi(fnames, sample_names=None):
    """Parse multiple bcf stats outputs and combine.

    :param fnames: list of filenames output by `bcftools stats.`
    :param samples_names: list of names of each sample, else the basename
        of each input file is used.
    """
    if sample_names is not None:
        if len(sample_names) != len(fnames):
            raise TypeError(
                "`fnames` and `sample_names` should be of equal length.")
    else:
        sample_names = [os.path.basename(f) for f in fnames]
    all_tables = defaultdict(list)
    for sample, fname in zip(sample_names, fnames):
        tables = parse_bcftools_stats(fname)
        # we don't use the 'id' field - its for intersections
        # and unions of two VCFs given to bcftools stats
        for k, t in tables.items():
            t.drop(columns='id', inplace=True)
            t.insert(0, 'sample', sample)
            all_tables[k].append(t)
    for k, t in all_tables.items():
        all_tables[k] = pd.concat(t)
    return dict(all_tables)  # don't want defaultdict



"""Report building from multiple items."""

from collections import OrderedDict
import textwrap
import uuid

from bokeh.embed import components
from bokeh.model import Model
from bokeh.models.sources import ColumnDataSource
from bokeh.models.widgets import DataTable, TableColumn
from bokeh.resources import INLINE
from jinja2 import Template
import markdown
import pkg_resources

JS_RESOURCES = [
    'simple-datatables_latest.js']


def _maybe_new_report(section, require_keys=False):
    """Create a new report section, or returns input."""
    if section is None:
        section = HTMLSection(require_keys=require_keys)
    else:
        if not isinstance(section, HTMLSection):
            raise TypeError("`section` should be an `HTMLSection`.")
    return section


class HTMLSection(OrderedDict):
    """A section of a report.

    The class can be extended o provide novel report components by
    overriding `._extra_components()`.
    """

    def __init__(self, require_keys=False):
        """Initialize the report item collection.

        :param title: report title.
        :param lead: report strapline, shown below title.
        :param require_keys: require keys when adding items.
        """
        self.require_keys = require_keys
        self.plots = list()
        self.md = markdown.Markdown()

    def _add_item(self, item, key=None):
        """Add an item to the report.

        :param item: item to add.
        :param key: unique key for item.
        """
        if key is None:
            if self.require_keys:
                raise ValueError('A key is required.')
            else:
                key = str(uuid.uuid4())
        self[key] = item

    def placeholder(self, key):
        """Add a placeholder to be filled in later."""
        self._add_item(None, key=key)

    def plot(self, plot, key=None):
        """Add a plot to the report.

        :param plot: bokeh plot instance.
        :param key: unique key for item.
        """
        self._add_item(plot, key=key)

    def table(
            self, data_frame, index=False, key=None, searchable=True,
            paging=True, sortable=True, **kwargs):
        """Add a pandas dataframe to the report.

        :param df: pandas dataframe instance.
        :param index: include dataframe index in output.
        :param key: unique key for item.
        :param shrink: shrink wrap the table to avoid whitespace.
        :param kwargs:
                       https://github.com/fiduswriter/Simple-DataTables/wiki/API
        """
        table = Table(
            data_frame, index=index, searchable=searchable,
            paging=paging, sortable=sortable, **kwargs)
        self._add_item(table, key=key)

    def markdown(self, text, key=None):
        """Add markdown formatted text to the report.

        :param text: markdown formatted text. The text will be dedented before
            use so it it safe to use triple-quoted strings indented to match
            code indentation.
        :param key: unique key for item.
        """
        if text is None or text == '':
            return
        text = textwrap.dedent(text)
        html = self.md.convert(text)
        self.md.reset()
        self._add_item(html, key=key)

    def alert(self, title, text, level, key=None):
        """Add an alert to the report.

        :param title: plain text - emphasis used on report
        :param text: plain text - the text of the warning
        :param level: the level of the warning; danger, warning, success, \
            or info
        :param key: unique key for item.
        """
        levels = ['danger', 'warning', 'success', 'info']
        if level not in levels:
            raise ValueError(
                'The alert level must be one of danger, warning, success, \
                    or info')

        if text is None or text == '':
            return

        html = '<div class="alert alert-'+level + \
            '"><p><strong>'+title+'</strong></p>'+text+'</div>'
        self._add_item(html, key=key)

    def _plot_components(self):
        """Return html script and div tags for bokeh plots."""
        # handle bokeh plots
        scripts = list()
        divs = dict()
        plots = {k: v for k, v in self.items() if isinstance(v, Model)}
        if len(plots) > 0:
            scripts, divs = components(plots)
            scripts = [scripts]
        return scripts, divs

    def _table_components(self):
        """Return html script and div for tables."""
        # TODO: reimplement `.table()` to not use bokeh?
        scripts = list()
        divs = dict()
        tables = {k: v for k, v in self.items() if isinstance(v, Table)}
        if len(tables) > 0:
            for k, v in tables.items():
                divs[k] = v.div
        return scripts, divs

    def _extra_components(self):
        """Return additional script and divs."""
        # hook for derived classes:
        # return: list of scripts, dict of divs keyed on self.keys
        return list(), dict()

    def components(self):
        """Fetch script and div tags for report."""
        scripts = list()
        specials = dict()
        # retrieve bokeh plot divs
        plot_scripts, plot_divs = self._plot_components()
        scripts.extend(plot_scripts)
        specials.update(plot_divs)
        # TODO: add special handing for new tables?
        table_scripts, table_divs = self._table_components()
        scripts.extend(table_scripts)
        specials.update(table_divs)
        # retrieve any extra specials
        extra_scripts, extra_divs = self._extra_components()
        scripts.extend(extra_scripts)
        specials.update(extra_divs)

        # put things together in order
        divs = list()
        for k in self.keys():
            try:
                divs.append(specials[k])
            except KeyError:
                if self[k] is None:
                    raise ValueError(
                        "Placeholder `{}` was not assigned "
                        "a value.".format(k)
                    )
                # fall back to putting items in directly
                divs.append(self[k])
        return scripts, divs


class HTMLReport(HTMLSection):
    """Generate HTML Report from a series of bokeh figures.

    Items added to the report take an optional key argument, adding items
    with the same key allows an update in place whilst maintaining the order
    in which items were added. Items can be grouped into sections for easier
    out of order addition.
    """

    def __init__(self, title="", lead="", require_keys=False, style='ont'):
        """Initialize the report item collection.

        :param title: report title.
        :param lead: report strapline, shown below title.
        :param require_keys: require keys when adding items.
        """
        super().__init__(require_keys=require_keys)
        self.title = title
        self.lead = lead
        self.sections = OrderedDict()
        self.sections['main'] = self
        self.style = style

        self.CSS_RESOURCES = dict(
                bootstrap='bootstrap.min.css',
                datatables='simple-datatables_latest.css',
                ont='custom-ont.css',
                ond='custom-ond.css',
                epi2me='custom-epi2me.css'
        )

        template = pkg_resources.resource_filename(
            __package__, 'data/report_template.html')
        with open(template, 'r', encoding="UTF-8") as fh:
            template = fh.read()
        self.template = Template(template)

    def add_section(self, key=None, section=None, require_keys=False):
        """Add a section (grouping of items) to the report.

        :param key: unique key for section.
        :param section: `HTMLSection` to add rather than creating anew.

        :returns: the report section.

        """
        if key is None:
            key = str(uuid.uuid4())
        section = _maybe_new_report(section, require_keys=require_keys)
        self.sections[key] = section
        return self.sections[key]

    def render(self):
        """Generate HTML report containing figures."""
        bokeh_resources = INLINE.render()

        libs = []
        css_files = ['bootstrap', 'datatables', self.style]
        CSS_RESOURCES = [self.CSS_RESOURCES[file] for file in css_files]
        for resources, stub in (
                [CSS_RESOURCES, "<style>{}</style>"],
                [JS_RESOURCES, '<script type="text/javascript">{}</script>']):
            for res in resources:
                fn = pkg_resources.resource_filename(
                    __package__, 'data/{}'.format(res))
                with open(fn, encoding="UTF-8") as fh:
                    libs.append(stub.format(fh.read()))

        all_scripts = list()
        all_divs = list()
        for sec_name, section in self.sections.items():
            scripts, divs = section.components()
            all_scripts.extend(scripts)
            all_divs.extend(divs)

        script = '\n'.join(all_scripts)
        divs = '\n'.join(all_divs)
        return self.template.render(
            title=self.title, lead=self.lead,
            bokeh_resources=bokeh_resources, resources="\n".join(libs),
            script=script, div=divs)

    def write(self, path):
        """Write html report to file."""
        with open(path, "w", encoding='utf8') as outfile:
            outfile.write(self.render())


class WFReport(HTMLReport):
    """Report template for epi2me-labs/wf* workflows."""

    def __init__(
            self, title, workflow, commit=None, revision=None,
            require_keys=False, about=True, style='ont', lead=None):
        """Initialize the report item collection.

        :param workflow: workflow name (e.g. wf-hap-snps)
        :param title: report title.
        :param require_keys: require keys when adding items.
        """
        if commit is None:
            commit = "unknown"
        if revision is None:
            revision = "unknown"
        self.wf_commit = commit
        self.wf_revision = revision
        self.workflow = workflow
        self.about = about
        self.style = style

        if lead is None:
            lead = (
                "Results generated through the {} Nextflow workflow "
                "provided by Oxford Nanopore Technologies.".format(workflow))

        super().__init__(
            title=title, lead=lead, require_keys=require_keys, style=style)
        self.tail_key = str(uuid.uuid4())

    def render(self):
        """Generate HTML report containing figures."""
        # delete and re-add the tail (in case we are called twice,
        # and something was added).
        try:
            del self.sections[self.tail_key]
        except KeyError:
            pass

        if self.about:
            self.add_section(key=self.tail_key).markdown("""
    ### About

    This report was produced using the
    [epi2me-labs/{0}](https://github.com/epi2me-labs/{0}).  The
    workflow can be run using `nextflow epi2me-labs/{0} --help`

    **Version details** *Revision*: {1} *Git Commit*: {2}


    ***Oxford Nanopore Technologies products are not intended for use for
    health assessment or to diagnose, treat, mitigate, cure or prevent any
    disease or condition.***


    ---
    """.format(self.workflow, self.wf_revision, self.wf_commit))
        return super().render()


def bokeh_table(df, index=True, **kwargs):
    """Generate a bokeh table from a pandas dataframe."""
    columns = [TableColumn(field=x, title=x) for x in df.columns]
    if index:
        columns = [TableColumn(field="index", title="Feature")] + columns
    return DataTable(
        columns=columns, source=ColumnDataSource(df), index_position=None,
        **kwargs)


class Table():
    """A table report component."""

    def __init__(
            self,
            data_frame, index, th_color='#0084A9', escape=True, **kwargs):
        """Initialize table component.

        :param dataframe: dataframe to turn in to simple table.
        """
        template = Template(
            """\
            <style>
            #{{ table_id }}{
            font-family: Arial, Helvetica, sans-serif;
            border-collapse: collapse;
            width: 100%;
            }
            #{{ table_id }} td, #{{ table_id }} th {
            border: 1px solid #ddd;
            padding: 4px;
            }
            #{{ table_id }} tr:nth-child(even){background-color: #f2f2f2;}
            #{{ table_id }} tr:hover {background-color: #90C5E7;}
            #{{ table_id }} th {
            padding-top: 4px;
            padding-bottom: 4px;
            text-align: left;
            background-color: """+th_color+""" ;
            color: white;
            }
            </style>

                    <body>
                        {{ dataframe }}
                    </body>
                   <script defer type="text/javascript">
                        let {{ table_id|safe }}=new simpleDatatables.DataTable(
                            "#{{ table_id }}",{
                         {% for k, v in kwargs.items() %}
                            {{ k }}: {{ v }},
                        {% endfor %}
                        });
                    </script>

            """)
        for key, val in kwargs.items():
            if isinstance(val, bool):
                kwargs[key] = str(val).lower()
        key = str(uuid.uuid4()).replace('-', '')
        key = "a{}".format(key[1:])
        # html id must not begin with numeric

        self.div = template.render(
            dataframe=data_frame.to_html(
                table_id=key, index=index, escape=escape),
            table_id=key, kwargs=kwargs)



"""Bio specific plots, could be generalised."""

from io import StringIO

from bokeh.models import Range1d
from bokeh.models.tickers import FixedTicker
from bokeh.plotting import figure
import numpy as np
import pandas as pd

from aplanat import util

_chrom_data_ = """
chrom length
1 249250621
2 243199373
3 198022430
4 191154276
5 180915260
6 171115067
7 159138663
8 146364022
9 141213431
10 135534747
11 135006516
12 133851895
13 115169878
14 107349540
15 102531392
16 90354753
17 81195210
18 78077248
19 59128983
20 63025520
21 48129895
22 51304566
X 155270560
Y 59373566
"""
_chrom_data_ = pd.read_csv(StringIO(_chrom_data_), delimiter=" ")


@util.plot_wrapper
def karyotype(
        x_datas, y_datas, names=None, colors=None, alpha=0.2,
        chrom_data=_chrom_data_, **kwargs):
    """Create a heatmap from three columns.

    :param x: x-axis coordinates (chromosome position)
    :param y: y-axis coordinates (chromosome)
    :param name: title for z-axis.
    :param colors: colours to plot tics for each data series.
    :param alpha: alpha level (transparency) with which to plot tics.
    :params chrom_data: a dataframe (or ndarray) with fields
        `chrom`, and `length` used to plot chromosome outline
        rectangles. Chromosomes will be plotted vertically (top to bottom)
        in the order given by this input.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    """
    if names is None:
        names = [None] * len(x_datas)
    if colors is None:
        colors = [None] * len(x_datas)
    defaults = {
        "output_backend": "webgl",
        "height": 300, "width": 600}
    defaults.update(kwargs)
    p = figure(**defaults)
    width = 0.8
    chrom_order = pd.Series(range(1, len(chrom_data) + 1))
    p.rect(
        chrom_data["length"] // 2, chrom_order,
        chrom_data["length"], width,
        fill_color='white', line_color='black')

    plot_order = dict(zip(chrom_data['chrom'], chrom_order))
    for x, y, name, color in zip(x_datas, y_datas, names, colors):
        y = np.array([plot_order[i] for i in y])
        kw = {'alpha': alpha}
        if name is not None:
            kw['legend_label'] = name
        if color is not None:
            kw['color'] = color
        p.segment(
            x, y - width / 2, x, y + width / 2, **kw)

    # set up the axes
    p.xaxis.formatter.use_scientific = False
    p.yaxis.ticker = FixedTicker()
    p.yaxis.ticker.ticks = chrom_order
    p.yaxis.major_label_overrides = {v: k for k, v in plot_order.items()}
    p.x_range = Range1d(
        start=0, end=max(chrom_data['length']),
        bounds=(0, max(chrom_data['length'])))
    miny, maxy = 0, len(chrom_data) + 1
    p.y_range = Range1d(
        start=miny, end=maxy, bounds=(miny, maxy))
    return p



"""Plots which are more graphics than plots."""

import os
import tempfile

from bokeh.layouts import gridplot
from bokeh.models import Range1d
from bokeh.models.annotations import Label
from bokeh.plotting import figure
from icon_font_to_png.icon_font import IconFont
import numpy as np
from PIL import Image
import pkg_resources
from si_prefix import si_format
import sigfig


class IconRGBA:
    """Wrapper to icon_font_to_png for multiple ttfs."""

    def __init__(self):
        """Initialize the class."""
        self.css_file = pkg_resources.resource_filename(
            __package__, 'data/fontawesome.css')
        self.ttf_files = [
            pkg_resources.resource_filename(__package__, 'data/{}'.format(x))
            for x in ('fa-regular-400.ttf', 'fa-solid-900.ttf')]
        self.icon_sets = [
            IconFont(self.css_file, ttf) for ttf in self.ttf_files]
        self.icons = dict()
        for i, icon_set in enumerate(self.icon_sets):
            self.icons.update(
                {k: (i, icon_set) for k in icon_set.css_icons.keys()})
        self.tmpdir = tempfile.TemporaryDirectory()

    def rgba(self, icon, size, color='black', scale='auto', uint32=True):
        """Create RGBA array for icon.

        :param icon: valid icon name
        :param size: icon size in pixels
        :param color: color name or hex value
        :param scale: scaling factor between 0 and 1,
            or 'auto' for automatic scaling
        :param uint32: return a two-dimension uint32-packed array (as
            required by bokeh image_rgba).
        """
        fname = icon + '.png'
        tmpfile = os.path.join(self.tmpdir.name, fname)
        try:
            icon_set = self.icon_sets[self.icons[icon][0]]
        except KeyError:
            raise KeyError('Unknown icon: {}.'.format(icon))
        icon_set.export_icon(
            icon, size, color, scale, filename=fname,
            export_dir=self.tmpdir.name)
        if not os.path.isfile(tmpfile):
            raise RuntimeError('Image not produced.')
        image = np.ascontiguousarray(
            np.array(Image.open(tmpfile))[::-1, :, :])
        if uint32:
            image = image.view(dtype=np.uint32).reshape(size, size)
        return image


fa_icons = IconRGBA()


class InfoGraphItems(dict):
    """Helper class to cumulatively create items for and infographic."""

    def __init__(self, *args, **kwargs):
        """Initialize the helper, and bootstrap fontawesome."""
        super().__init__(*args, **kwargs)

    def append(self, label, value, icon, unit=''):
        """Add an item.

        :param label: infographic item label.
        :param value: numerical value of headline number (without SI units).
        :param icon: font-awesome icon to use.
        :param unit: additional suffix after SI unit suffix, e.g. "bases".

        """
        try:
            fa_icons.icons[icon]
        except KeyError:
            raise KeyError("'{}' is not a known icon.".format(icon))
        self[label] = (label, value, icon, unit)

    def extend(self, items):
        """Add multiple items at once.

        :param items: iterable of 4-tuples, as required by `.add()`.
        """
        for i in items:
            self.append(*i)


def infographic(items, **kwargs):
    """Create and infographic 'plot'.

    :param items: 3-tuples of (label, value, unit, icon); the label should be
        a one or two word description, the value the headline number, and the
        icon the name of a fontawesome icon. If `value` is  numeric, it
        will be normalised by use of an SI suffix for display after which
        `unit` will be appended. If `value` is a string it will be used as is.
    :param kwargs: kwargs for bokeh gridplot.

    ..note:: If `bootstrap_fontawesome` has not already been called, the
        function will load the required fonts, however they will not display
        the first time an Jupyter labs cell is run. If using the
        `InfoGraphItems` helper class, this wrinkle will be taken care of
        provided the helper is initiated in a previous cell.

    """
    plots = list()
    seen = set()
    for label, value, icon, unit in items:
        if label in seen:
            continue
        if not isinstance(value, str):
            if unit == '%':
                value = "{}%".format(sigfig.round(value, 3))
            else:
                value = si_format(value) + unit
        seen.add(label)
        width, height = 175, 100
        aspect = height / width
        p = figure(
            output_backend='webgl',
            plot_width=width, plot_height=height,
            title=None, toolbar_location=None)
        p.axis.visible = False
        p.grid.visible = False
        p.outline_line_color = None
        p.rect([0.5], [0.5], [1.0], [1.0], fill_color="#2171b5")
        p.x_range = Range1d(start=0.1, end=0.9, bounds=(0.1, 0.9))
        p.y_range = Range1d(start=0.1, end=0.9, bounds=(0.1, 0.9))
        p.add_layout(
            Label(
                x=0.15, y=0.45, text=value, text_color="#DEEBF7",
                text_font_size="24px"))
        p.add_layout(
            Label(
                x=0.15, y=0.2, text=label, text_color="#C6DBEF",
                text_font_size="16px"))
        image = fa_icons.rgba(icon, 75, color='#6BAED6')
        p.image_rgba(image=[image], x=0.6, y=0.4, dw=0.25, dh=0.25/aspect)
        plots.append(p)
    defaults = {'toolbar_location': None, "ncols": len(items)}
    defaults.update(kwargs)
    return gridplot(plots, **defaults)



#!/usr/bin/env python
"""Create depth coverage report."""

import argparse

from bokeh.layouts import gridplot, layout
from bokeh.models import Panel, Tabs
import numpy as np
import pandas as pd

from aplanat import lines
from aplanat.report import _maybe_new_report, HTMLReport
from aplanat.util import Colors


_full_report_header = """
### Depth Coverage

The following tables and figures are derived from
the output of [Mosdepth]
(https://github.com/brentp/mosdepth).
"""


def cumulative_depth_from_dist(depth_file: str, **kwargs):
    """Cumulative depth plots from mosdepth dist file.

    :param: depth_file: mosdepth.*.dist.txt file
    """
    df = pd.read_csv(
        depth_file, sep='\t', names=['ref', 'coverage', 'proportion'])
    # Use whole genome: 'total'
    df = df[df.ref == 'total']
    df.sort_values('coverage', ascending=True, inplace=True)
    df.proportion = df.proportion * 100
    df.drop_duplicates('proportion', inplace=True)

    p = lines.line(
        [df.coverage], [df.proportion],
        x_axis_label='Read depth',
        y_axis_label='Percentage of genome',
        **kwargs)
    return p


def cumulative_depth_from_bed(df: pd.DataFrame, bins: int = 2000, **kwargs):
    """Cumulative depth plot from a mosdepth bed derived dataframe.

    :param df: depth dataframe
        Required columns:
        - start
        - end
        - depth
    :param: bins: number of bins to plot
    :param: kwargs: keyword arguments for aplanat.lines.line
    """
    # Count bases covered by each "step" in the BED
    df['step'] = df.end - df.start

    # * Merge steps with the same depth together for total per-depth base count
    # * Sort descending to count cumulatively as coverage decreases
    #   ie. the proportion of counted bases approaches 1 as we reach 0 cov
    df_agg = df.groupby("depth") \
        .agg({"step": "sum"}) \
        .sort_values("depth", ascending=False)
    df_agg["step_cumsum"] = df_agg.step.cumsum()
    df_agg["percent_at_depth"] = \
        df_agg.step_cumsum / df_agg.step_cumsum.max() * 100

    # Flip table to correct axes
    df_agg = df_agg[::-1]

    # Select slices (depths are accessed by .index)
    x = df_agg.index.to_numpy()
    y = df_agg.percent_at_depth.to_numpy()
    if len(x) > bins:
        binner = np.linspace(0, len(df_agg) - 1, bins).astype(int)
        x_bin = np.array(x)[binner]
        y_bin = np.array(y)[binner]
    else:
        x_bin = x
        y_bin = y

    p = lines.line(
        [x_bin], [y_bin],
        x_axis_label='Read depth',
        y_axis_label='Percentage of genome',
        **kwargs)
    return p


def depth_coverage(depth_file, xlim=(None, None), ylim=(None, None), **kwargs):
    """Create plot of depth coverage by region per ref name.

    :param depth_file: depth file output from mosdepth
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param ylim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.

    :returns: a list of bokeh plots.
    """
    depth_file = pd.read_csv(depth_file, sep='\t')
    depth_file.columns = ['ref', 'start', 'end', 'depth']
    all_ref = dict(tuple(depth_file.groupby(['ref'])))
    plots = []
    for ref in sorted(all_ref):
        depths = all_ref[ref]
        plot = lines.steps(
            list([depths['start']]), list([depths['depth']]),
            colors=[Colors.cerulean], mode='after',
            x_axis_label='Position along reference',
            y_axis_label='Sequencing depth / Bases',
            title=str(ref), xlim=xlim, ylim=ylim, **kwargs)
        plot.xaxis.formatter.use_scientific = False
        plots.append(plot)
    return plots


def depth_coverage_orientation(
        fwd, rev, xlim=(None, None), ylim=(None, None), **kwargs):
    """Create plot of depth coverage by region per ref name with fwd and rev.

    :param fwd: fwd depth file output from mosdepth
    :param rev: rev depth file output from mosdepth
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param ylim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.

    :returns: a list of bokeh plots.
    """
    depth_file = pd.read_csv(fwd, sep='\t')
    depth_file.columns = ['ref', 'start', 'end', 'fwd']
    rev_file = pd.read_csv(rev, sep='\t')
    rev_file.columns = ['ref', 'start', 'end', 'rev']
    depth_file['rev'] = rev_file['rev']
    all_ref = dict(tuple(depth_file.groupby(['ref'])))
    plots = []
    for ref in sorted(all_ref):
        depths = all_ref[ref]
        plot = lines.steps(
            [list(depths['start']), list(depths['start'])],
            [list(depths['fwd']), list(depths['rev'])],
            colors=[Colors.cerulean, Colors.feldgrau],
            names=['fwd', 'rev'], mode='after',
            x_axis_label='Position along reference',
            y_axis_label='Sequencing depth / Bases',
            title=str(ref), xlim=xlim, ylim=ylim, **kwargs)
        plot.xaxis.formatter.use_scientific = False
        plots.append(plot)
    return plots


def full_report(
        depth_file, fwd, rev, header=_full_report_header, report=None,
        sample_counts=False,
        tab=False, **kwargs):
    """Create a report section from the output of fastcat.

    :param depth_file: a depth file outout from mosdepth.
    :param fwd: fwd depth file output from mosdepth
    :param rev: rev depth file output from mosdepth
    :param header: a markdown formatted header.
    :param report: an HTMLSection instances
    :param tab: tabular output

    :returns: an HTMLSection instance, if `report`
        was provided the given instance is modified and returned.
    """
    report = _maybe_new_report(report)
    report.markdown(header)

    plots_coverage = depth_coverage(depth_file)
    plots_orient = depth_coverage_orientation(fwd, rev)

    if tab:
        tab1 = Panel(
                child=gridplot(plots_coverage, ncols=1),
                title="Proportions covered")
        tab2 = Panel(
                child=gridplot(plots_orient, ncols=1),
                title="Coverage traces")
        plots = Tabs(tabs=[tab1, tab2])
        report.plot(plots)
    else:
        plots = [[plots_coverage, plots_orient]]
        report.plot(layout(plots, sizing_mode="stretch_width"))
    return report


def main(args):
    """Entry point to create a report from depth file."""
    report = full_report(
        args.depth_file, args.fwd, args.rev, report=HTMLReport(),
        tab=args.tab)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        "Depth coverage for one input file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--depth_file",
        help="Mosdepth depth file.")
    parser.add_argument(
        "--fwd",
        help="Mosdepth fwd file.")
    parser.add_argument(
        "--rev",
        help="Mosdepth rev file.")
    parser.add_argument(
        "--tab", default=False,
        help="Tabular output")
    parser.add_argument(
        "--output", default="depth_coverage.html",
        help="Output HTML file.")

    return parser



"""Report components for displaying information from bcftools stats."""

import argparse

from bokeh.layouts import layout
import numpy as np
import pandas as pd

from aplanat import annot, bars, hist
from aplanat.report import _maybe_new_report, HTMLReport
from aplanat.util import Colors, read_files


def read_length_plot(
        seq_summary, min_len=None,
        max_len=None, xlim=(0, None)):
    """Create a read length plot.

    :param seq_summary: summary data from fastcat.
    :param min_len: minimum length.
    :param max_len: maximum length.
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.

    The minimum and maximum lengths are used only to annotate the plot
    (not filter the data).
    """
    total_bases = seq_summary['read_length'].sum()
    try:
        mean_length = total_bases / len(seq_summary)
    except ZeroDivisionError:
        mean_length = 0
    median_length = np.median(seq_summary['read_length'])
    datas = [seq_summary['read_length']]
    length_hist = hist.histogram(
        datas, colors=[Colors.cerulean], bins=100,
        title="Read length distribution.",
        x_axis_label='Read Length / bases',
        y_axis_label='Number of reads',
        xlim=xlim)
    if min_len is not None:
        length_hist = annot.marker_vline(
            length_hist, min_len,
            label="Min: {}".format(min_len), text_baseline='bottom',
            color='grey')
    if max_len is not None:
        length_hist = annot.marker_vline(
            length_hist, max_len,
            label="Max: {}".format(max_len), text_baseline='top')
    length_hist = annot.subtitle(
        length_hist,
        "Mean: {:.0f}. Median: {:.0f}".format(
            mean_length, median_length))
    return length_hist


def read_quality_plot(seq_summary):
    """Create read quality plot.

    :param seq_summary: summary data from fastcat.
    """
    datas = [seq_summary['mean_quality']]
    mean_q, median_q = np.mean(datas[0]), np.median(datas[0])
    q_hist = hist.histogram(
        datas, colors=[Colors.cerulean], bins=100,
        title="Read quality score",
        x_axis_label="Quality score",
        y_axis_label="Number of reads",
        xlim=(4, 25))
    q_hist = annot.subtitle(
        q_hist,
        "Mean: {:.0f}. Median: {:.0f}".format(
            mean_q, median_q))
    return q_hist


def sample_read_counts(seq_summary, min_len=None, max_len=None):
    """Create bar plot counting unique samples.

    :param seq_summary: summary data from fastcat.
    :param min_len: minimum length.
    :param max_len: maximum length.
    """
    if min_len is not None:
        seq_summary = seq_summary.loc[
            (seq_summary['read_length'] > min_len)]
    if max_len is not None:
        seq_summary = seq_summary.loc[
            (seq_summary['read_length'] < max_len)]
    sample_counts = (
        pd.DataFrame(seq_summary['sample_name'].value_counts())
        .sort_index()
        .reset_index()
        .rename(
            columns={'index': 'sample', 'sample_name': 'count'})
        )

    title = 'Number of reads per barcode'
    if min_len is not None or max_len is not None:
        t0, t1 = "", ""
        if min_len is not None:
            t0 = "{} < ".format(min_len)
        if max_len is not None:
            t1 = " > {}".format(max_len)
        title += " (filtered by {}length{}).".format(t0, t1)

    plot = bars.simple_bar(
        sample_counts['sample'].astype(str), sample_counts['count'],
        colors=[Colors.cerulean]*len(sample_counts),
        title=title,
        plot_width=None)
    plot.xaxis.major_label_orientation = 3.14/2
    return plot


_full_report_header = """
### Read summary

The following tables and figures are derived from
the output of [fastcat](https://github.com/epi2me-labs/fastcat).
"""


def full_report(
        stats, header=_full_report_header, report=None,
        sample_counts=False, min_len=None, max_len=None):
    """Create a report section from the output of fastcat.

    :param stats: one or more outputs from `fastcat`.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.
    :param sample_counts: include barplot of sample counts.
    :param max_len: maximum read length.
    :param max_len: minimum read length.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    stats = read_files(stats)
    report = _maybe_new_report(report)
    report.markdown(header)

    read_length = read_length_plot(stats, min_len=min_len, max_len=max_len)
    read_qual = read_quality_plot(stats)
    plots = [[read_length, read_qual]]
    if sample_counts:
        plots.append(
            [sample_read_counts(stats, min_len=min_len, max_len=max_len)])
    report.plot(layout(plots, sizing_mode="stretch_width"))

    return report


def main(args):
    """Entry point to create a report from bcftools stats."""
    report = full_report(
        args.stats_files, report=HTMLReport(),
        min_len=args.min_len, max_len=args.max_len,
        sample_counts=args.samples)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        "fastcat per-read report",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--samples", action='store_true',
        help="Show sample counts.")
    parser.add_argument(
        "--min_len", type=int,
        help="Minimum length filter.")
    parser.add_argument(
        "--max_len", type=int,
        help="Maximum length filter.")
    parser.add_argument(
        "--output", default="fastcat_report.html",
        help="Output HTML file.")
    parser.add_argument(
        "stats_files", nargs='+',
        help=(
            "One or more files containing the per-read output from "
            "`fastcat`."))
    return parser



"""Creation of more complex plotting layouts."""

import itertools

from bokeh.layouts import gridplot, layout, Row, Spacer
from bokeh.models import Range1d
from bokeh.models.annotations import Label
from bokeh.plotting import figure

from aplanat import util


def facet_grid(
        df, aes, plot_func,
        xlim=(None, None), ylim=(None, None),
        facet=(None, None),
        x_axis_label="", y_axis_label="",
        x_facet_heading="", y_facet_heading="",
        transform=(None, None), link_axes=True,
        **kwargs):
    """Mainly for the creation of facet-grid plots from a Dataframe.

    :param df: pandas dataframe
    :param aes: a dictionary with keys 'x', 'y', 'col' corresponding to the
        data to plot along the x-axis, y-axis, and optional colouring. These
        should be column names of the input dataframe.
    :param plot_func: aplanat plotting function to use.
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param ylim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param facet: a tuple indicating the variables by which to create (x, y)
        facets of the data.
    :param x_axis_label: label text for the x axis.
    :param y_axis_label: label text for the y axis
    :param x_facet_heading: facet name for the columns (x facet)
    :param y_facet_heading: facet name for the rows (y facet)
    :param transform: a tuple of functions to apply to the (x, y) data of each
        facet before plotting.
    :param link_axes: link axes of subplots such that panning and zooming
        effects all plots.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh layout.
    """
    if 'x' not in aes or 'y' not in aes:
        raise KeyError("'aes' argument must contain keys: 'x', 'y'.")
    # find unique values of facets
    facet_x, facet_y = facet
    facet_x_values, facet_y_values = [None], [None]
    # and colours
    col_values = [None]
    if 'col' in aes:
        col_values = df[aes['col']].unique()
        cols = util.choose_palette(len(col_values))
        cols_dict = dict(zip(col_values, cols))

    if facet_x is not None:
        facet_x_values = df[facet_x].unique()
    if facet_y is not None:
        facet_y_values = df[facet_y].unique()
    facet_products = list(itertools.product(
        enumerate(facet_x_values), enumerate(facet_y_values)))

    plots = list()
    for (px, fx), (py, fy) in facet_products:
        d = df
        if fx is not None and fy is not None:
            d = df.loc[(df[facet_x] == fx) & (df[facet_y] == fy)]
        elif fx is not None:
            d = df.loc[df[facet_x] == fx]
        elif fy is not None:
            d = df.loc[df[facet_y] == fy]

        x_data = list()
        y_data = list()
        colors = list()
        for col in col_values:
            if col is not None:
                dcol = d.loc[d[aes['col']] == col]
            else:
                dcol = d
            if len(dcol) == 0:
                continue
            colors.append(cols_dict[col])
            # create plot vectors with transform if required
            x = dcol[aes['x']]
            if transform[0] is not None:
                x = transform[0](x)
            x_data.append(x)
            if 'y' in aes:
                y = dcol[aes['y']]
                if transform[1] is not None:
                    y = transform[1](y)
                y_data.append(y)
        # make the plot
        if len(x_data) == 0:
            continue  # no data for this facet
        plot = plot_func(
            x_data, y_data, colors=colors,
            xlim=xlim, ylim=ylim,
            **kwargs)
        plots.append(((px, py), plot))

    # fix limits of all plots to the same
    for attr in ['x_range', 'y_range']:
        zlims = [
            (getattr(p, attr).start, getattr(p, attr).end)
            for _, p in plots]
        zmin, zmax = min(z[0] for z in zlims), max(z[1] for z in zlims)
        zrange = Range1d(start=zmin, end=zmax, bounds=(zmin, zmax))
        for _, p in plots:
            if not link_axes:
                # nb: copy.deepcopy doesn't do as intended
                zrange = Range1d(start=zmin, end=zmax, bounds=(zmin, zmax))
            setattr(p, attr, zrange)

    # make the main panel of plots
    ncols, nrows = len(facet_x_values), len(facet_y_values)
    grid = []
    for _ in range(nrows):
        grid.append([None]*ncols)
    for (col, row), p in plots:
        grid[row][col] = p
    plotgrid = gridplot(grid)
    # figure out the height and width that were used
    bh, bw = plots[0][1].plot_height, plots[0][1].plot_width

    # get labels and headers
    x_header, x_axis_label, *x_labels = make_facet_labels(
        x_facet_heading, facet_x_values, x_axis_label,
        base_height=bh, base_width=bw)
    y_header, y_axis_label, *y_labels = make_facet_labels(
        y_facet_heading, facet_y_values, y_axis_label, facet='y',
        base_height=bh, base_width=bw)

    # add legend
    # bokeh doesn't support legends outside a figure
    # https://github.com/bokeh/bokeh/issues/7607
    # so make a whole new figure
    leg_plot = plot_func(
        [[0, 0.1]]*len(col_values), [[0, 0.1]]*len(col_values),
        colors=cols, names=col_values,
        title=None, toolbar_location=None,
        height=bh*len(facet_y_values), width=150)
    leg_plot.axis.visible = False
    leg_plot.grid.visible = False
    leg_plot.outline_line_color = None
    # fudge plot to be outside where data is
    leg_plot.x_range = Range1d(start=-1.0, end=-0.5)
    leg_plot.y_range = Range1d(start=-1.0, end=-0.5)

    # create the plot layout, row by row.
    plot_layout = []
    # toobox, aligned right
    plot_layout.append(
        [Spacer(sizing_mode='stretch_width'), plotgrid.children[0]])
    if len(facet_x_values) > 1:
        # x facet header
        plot_layout.append(
            [Spacer(width=40), x_header, Spacer(sizing_mode='stretch_width')])
        # x facet labels
        plot_layout.append(
            [Spacer(width=40)] + x_labels +
            [Spacer(sizing_mode='stretch_width')])
    # y axis label, plots, y facet labels, y facet header
    row = [
        y_axis_label, Row(plotgrid.children[1])]
    if len(facet_y_values) > 1:
        row.extend([y_labels, y_header])
    if len(col_values) > 1:
        row.append(leg_plot)
    plot_layout.append(row)
    # x axis label
    plot_layout.append(
        [Spacer(width=40), x_axis_label, Spacer(sizing_mode='stretch_width')])
    # make plot
    plot = layout(plot_layout)
    return plot


def make_facet_labels(
        facet_name, facet_labels, axis_label,
        facet='x', base_height=300, base_width=400):
    """Create labels for a facet grid plot.

    :param facet_name: the header name for the facet.
    :param facet_labels: the names of the factors of the facet.
    :param axis_label: the axis label (as would appear on a single plot)
    :param facet: either 'x' or 'y'.
    :param base_height: the height of the corresponding plots.
    :param base_width: the width of the corresponding plots.

    :returns: [facet header, axis label, facet 1 label, ...]

    """
    plots = list()
    h, w = 40, base_width
    head_h, head_w = 40, len(facet_labels) * w
    angle = 0
    axis_angle = 0
    plots = list()
    if facet == 'y':
        h, w = base_height, 40
        head_h, head_w = len(facet_labels) * h, 40
        angle = -3.14/2
        axis_angle = 3.14/2

    def _label_plot(label, h, w, angle=angle, **kwargs):
        # a figure that simply holds a label
        p = figure(
            output_backend='webgl',
            plot_height=h, plot_width=w,
            title=None, toolbar_location=None)
        p.axis.visible = False
        p.grid.visible = False
        p.outline_line_color = None
        p.toolbar.active_drag = None
        p.x_range = Range1d(start=0.0, end=1.0)
        p.y_range = Range1d(start=0.0, end=1.0)
        if label is None:
            label = ""
        p.add_layout(
            Label(
                x=0.5, y=0.5, text=label, text_align='center',
                text_baseline='middle', angle=angle, **kwargs))
        return p

    # create header and axis label
    plots.append(_label_plot(facet_name, h=head_h, w=head_w))
    plots.append(
        _label_plot(
            axis_label, h=head_h, w=head_w, angle=axis_angle,
            text_font_style='italic'))
    # label for each facet (column/row)
    for label in facet_labels:
        plots.append(_label_plot(label, h=h, w=w))
    return plots



"""Aplanat demo."""

import argparse
from copy import deepcopy

from bokeh.layouts import gridplot
import numpy as np
import pandas as pd

import aplanat
from aplanat import \
    annot, bars, bio, graphics, hist, lines, points, spatial, util
from aplanat.report import HTMLReport


def main(args):
    """Aplanat demo entry point."""
    util.set_basic_logging()
    logger = util.get_named_logger("Aplanat Demo")

    x = np.random.normal(size=2000)
    y = np.random.normal(size=2000)
    sorted_xy = [np.sort(x), np.sort(x)]

    # Start a report
    logger.info("Start a report")
    report = HTMLReport(
        title="Aplanat Demo",
        lead="A brief demonstration of Aplanat's API.",
        require_keys=False)  # set True to require keys on item addition

    # The report is an ordered dictionary, so we can add placeholders to
    # delay the addition of items. Markdown can be used for text items.
    report.markdown("placeholder", key="simple preamble")

    # We can also add additional sections. These function as their own
    # mini-report and each section is rendered in its entirety in the order it
    # was added.  (The main HTMLReport is logically the first section).
    logger.info("Adding gallery")
    gallery = report.add_section(key="additional_section")

    # Adding a plot
    logger.info("Adding points plot")
    pnt_plot = points.points(sorted_xy, sorted_xy[::-1])
    report.plot(pnt_plot, key="points_plot")

    # Test we can output a plot to json in a different doc/layout
    # using embed.json_item() fails in this case
    logger.info("Testing json export")
    report.markdown("A simple line plot.")
    line_plot = lines.line(sorted_xy, sorted_xy[::-1])
    json_plot = aplanat.dump_json(line_plot)
    logger.info("Some json: {}...".format(json_plot[:40]))
    report.plot(gridplot([[line_plot]]), key="lines_plot")

    # Using placeholder is more explicit, and checks will be made before
    # rendering that the item has be assigned a real value
    report.placeholder("histogram preamble")

    # There's no need to provide key (unless require_keys is set). Items added
    # without a key cannot be replaced however.
    logger.info("Adding histogram")
    h = hist.histogram([x - 1, y + 1], colors=['red', 'green'])
    h = annot.marker_vline(
        h, np.mean(x) - 1, label='x values - 1', color='red',
        text_baseline='bottom')
    h = annot.marker_vline(
        h, np.mean(y) + 1, label='y values + 1', color='green',
        text_baseline='top')
    report.plot(h)

    # To delete an item, just delete the key
    logger.info("Deleting an item for fun")
    report.markdown("Garbage", key='garbage')
    del report['garbage']

    # Add more plots
    logger.info("Adding heatmap")
    report.placeholder("heatmap preamble")
    report.plot(spatial.heatmap2(x, y))

    # Add a data table
    logger.info("Adding table")
    report.placeholder("table preamble")
    df = pd.DataFrame({'x': x, 'y': y})
    report.markdown(
        "Here's a table:")
    df = df[0:5]
    report.table(
        df, key='Table with auto_height', pagination='false',
        sortable='false')
    report.markdown(
        "Here's a second table with more data, pagination, and searchable")
    df = pd.DataFrame({'x': y, 'y': x})
    report.table(df, key='Table with more data and pagination', height=200)

    # boxplot series with discretised x
    logger.info("Adding boxplot")
    report.placeholder("boxplot preamble")
    x_discrete = np.around(x, 0)
    x_str = [str(x) for x in np.abs(x_discrete)]
    report.plot(
        gridplot([
            bars.boxplot_series(x_discrete, y, width=300, title='continuous'),
            bars.boxplot_series(x_str, y, width=300, title='categorical')],
            ncols=2))

    # Gallery
    gallery.placeholder("gallery preamble")
    logger.info("Adding some infographics")
    exec_summary = graphics.InfoGraphItems()
    exec_summary.append('Example', 0.0051, 'angle-up', '%')
    exec_summary.append('Total reads', 1000000, 'angle-up', '')
    exec_summary.append('Total yield', 1e9, 'signal', 'b')
    exec_summary.append('Mean read length', 50e3, 'align-center', 'b')
    exec_summary.append('Mean qscore (pass)', 14, 'thumbs-up', '')
    plot = graphics.infographic(exec_summary.values())
    gallery.plot(plot)

    logger.info("Adding karyogram")
    chrom_data = deepcopy(bio._chrom_data_)
    chrom_data['chrom'] = chrom_data['chrom'].apply(lambda x: 'chr' + x)
    chroms = np.random.choice(chrom_data['chrom'], size=10000)
    positions = list()
    for chrom in chroms:
        length = chrom_data.loc[chrom_data['chrom'] == chrom, 'length']
        positions.append(np.random.randint(length)[0])
    plot = bio.karyotype([positions], [chroms], chrom_data=chrom_data)
    gallery.plot(plot)

    report.placeholder("failed plot preamble")
    logger.info(
        "Adding failed plot -- an exception trace will be logged below.")
    group = ("x1", "x2", "x3", 4, "x5")
    count = ('wrong data')

    @util.plot_wrapper
    def decorator_example(groups, counts):
        """Exemplify a decorator."""
        p = bars.simple_bar(groups, counts)
        return p

    plot = decorator_example(group, count)
    report.plot(plot)
    logger.info("End of plot that will fail.")

    # Trying to render now will raise ValueError because the placeholders are
    # not filled in
    logger.info("Checking empty pladeholder validation")
    try:
        report.render()
    except ValueError:
        logger.exception("Caught error as expected:")
    logger.info("End of expected exception log.")

    # Fill in the placeholders
    report.markdown("""
    ### Simple plots

    A simple points plot:
    """, "simple preamble")

    report.markdown("""
    ### Histograms

    Multi-variate histograms:
    """, "histogram preamble")

    report.markdown("""
    ### Heatmaps

    Heatmaps from `x-y` values:
    """, "heatmap preamble")

    report.markdown("""
    ### Tables

    Data tables:
    """, "table preamble")

    report.markdown("""
    ### Boxplots

    A series of boxplots using either a continuous or categorical grouping:
    """, "boxplot preamble")

    report.markdown("""
    ### Failed Plot

    When plot fails:
    """, "failed plot preamble")

    gallery.markdown("""
    ### Gallery

    Assortment of possibilities:
    """, "gallery preamble")

    # write the output, implicitely calls .render()
    logger.info("Rendering report")
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        "aplanat demo",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--output", default="aplanat_demo_report.html",
        help="Output HTML file.")
    return parser



"""Report component for displaying information from mapula's json output."""

import argparse
import json
import math
import sys
import typing

from bokeh.layouts import gridplot, layout
from bokeh.models import Div, HoverTool, Panel, Slope, Tabs, Title
from bokeh.palettes import Category20c
from bokeh.plotting import figure
from bokeh.transform import cumsum
import markdown
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

from aplanat import hist, points
from aplanat.report import HTMLReport, HTMLSection
from aplanat.util import Colors


class PlotMappingStats(HTMLSection):
    """Build an aplanat dashboard from mapula's json output."""

    def __init__(
        self,
        json: str,
        counts: typing.Union[str, None],
    ) -> None:
        """Load the json file and output the dashboard."""
        super().__init__()
        self.json = json
        self.md = markdown.Markdown()
        self.counts = counts

        self.data = self.load_data(self.json)
        self.build_report(
            self.counts,
            **self.data
        )

    @staticmethod
    def load_data(
        path: str,
    ) -> dict:
        """Load_data."""
        try:
            with open(path) as data:
                try:
                    return json.load(data)
                except json.decoder.JSONDecodeError:
                    print('Error loading data from {}'.format(
                        path))
                    raise
        except IOError:
            print("Path {} cannot be opened".format(path))
            raise

    def build_report(
        self,
        counts: pd.DataFrame,
        **data
    ):
        """Build_report."""
        tabs = []

        # Plot summary
        summary_tab = self.build_summary_tab(data)
        tabs.append(summary_tab)

        # Plot accuracy
        acc_tab = self.build_accuracy_tab(data)
        tabs.append(acc_tab)

        # Plot quality
        qual_tab = self.build_quality_tab(data)
        tabs.append(qual_tab)

        # Plot lengths
        len_tab = self.build_length_tab(data)
        tabs.append(len_tab)

        # Plot coverage
        cov_tab = self.build_coverage_tab(data)
        tabs.append(cov_tab)

        # Plot control
        if counts:
            con_tab = self.build_control_tab(data, counts)
            tabs.append(con_tab)

        # Tabula rasa
        panel = Tabs(tabs=tabs)
        self.plot(panel)

    def build_summary_tab(self, data):
        """Build_summary_tab."""
        alignment_plots = [
            self.plot_alignment_summary(value)
            for value in data.values()
            if value['primary_count']
        ]

        if not alignment_plots:
            return None

        alignment_grid = gridplot(
            alignment_plots, ncols=2,
            sizing_mode="stretch_width")

        text1 = (
            "This report contains visualisations of "
            "statistics that can help in understanding "
            "the results from the wf-alignment workflow. "
            "Each tab above contains a different type of "
            "plot, and in general the results are broken "
            "down by the reference genome to which "
            "alignments were made. On this tab, the donut "
            "plot displays the proportion of mapped to "
            "unmapped reads (hover over for counts) and "
            "the bar charts give a breakdown of the different "
            "alignment types per-reference."
        )
        plots = [
            [self.get_description(text1)],
            [self.plot_base_pairs(data)],
            [alignment_grid]
        ]

        main = layout(plots, sizing_mode="scale_width")
        return Panel(child=main, title="Summary")

    def build_accuracy_tab(self, data):
        """Build_accuracy_tab."""
        accuracy_plots = [
            self.plot_accuracy_distribution(value)
            for value in data.values()
            if sum(value['alignment_accuracies'])
        ]

        if not accuracy_plots:
            return None

        accuracy_grid = gridplot(
            accuracy_plots, ncols=2,
            sizing_mode="stretch_width")

        text = (
            "This tab contains visualisations of "
            "percentage alignment accuracy broken "
            "down by aligned reference. Naturally, "
            "no accuracy measurements are available "
            "for unmapped reads."
        )
        plots = [
            [self.get_description(text)],
            [accuracy_grid]
        ]

        main = layout(plots, sizing_mode="scale_width")
        return Panel(child=main, title="Accuracy")

    def build_coverage_tab(self, data):
        """Build_coverage_tab."""
        coverage_plots = [
            self.plot_coverage_distribution(value)
            for value in data.values()
            if sum(value['alignment_coverages'])
        ]

        if not coverage_plots:
            return None

        coverage_grid = gridplot(
            coverage_plots, ncols=2,
            sizing_mode="stretch_width")

        text = (
            "This tab contains visualisations of "
            "reference-coverage (i.e. the proportion of "
            "the reference that a given alignment covers "
            "ex. soft-clipping). The cov80 gives the number "
            "of alignments which cover more than 80 percent "
            "of the reference."
        )
        plots = [
            [self.get_description(text)],
            [coverage_grid]
        ]

        main = layout(plots, sizing_mode="scale_width")
        return Panel(child=main, title="Coverage")

    def build_length_tab(self, data):
        """Build_length_tab."""
        length_plots = [
            self.plot_read_length_distribution(value)
            for value in data.values()
            if sum(value['read_lengths'])
        ]

        if not length_plots:
            return None

        length_grid = gridplot(
            length_plots, ncols=2,
            sizing_mode="stretch_width")

        text = (
            "This tab contains visualisations of "
            "read length. The n50 is defined as the "
            "length N for which 50 percent of all "
            "bases in the sequence are in a sequence "
            "of length L < N."
        )
        plots = [
            [self.get_description(text)],
            [length_grid]
        ]

        main = layout(plots, sizing_mode="scale_width")
        return Panel(child=main, title="Length")

    def build_quality_tab(self, data):
        """Build_quality_tab."""
        quality_plots = [
            self.plot_qscore_distribution(value)
            for value in data.values()
            if sum(value['aligned_qualities'])
        ]

        if not quality_plots:
            return None

        quality_grid = gridplot(
            quality_plots, ncols=2,
            sizing_mode="stretch_width")

        text = (
            "This tab contains visualisations of "
            "avg read quality. Each data point "
            "represents the mean per-base Phred quality "
            "score for a given read. Quality scores are "
            "related to the underlying error probabilities. "
            "[Note]: Unmapped quality scores will be "
            "provided in an update. "
        )
        plots = [
            [self.get_description(text)],
            [quality_grid]
        ]

        main = layout(plots, sizing_mode="scale_width")
        return Panel(child=main, title="Quality")

    def build_control_tab(self, data, counts):
        """Build_control_tab."""
        counts_df = pd.read_csv(counts)
        counts_df.columns = map(str.lower, counts_df.columns)

        if not all(
            col in counts_df.columns
            for col in ['reference', 'expected_count']
        ):
            print(
                "[Error]: Supplied counts file does not "
                "contain the required columns, "
                "'reference,expected_count'"
            )
            sys.exit(1)

        counts_df = counts_df.set_index('reference')
        required = ['spearmans_rho', 'observed_references']
        text = (
            "This tab becomes available if you "
            "have supplied an expected counts csv to the "
            "wf-alignment workflow. Providing expected counts "
            "permits us to calculate the correlation between "
            "the number of observations for each of a given set "
            "of reference sequences against their expected counts. "
        )
        plots = [
            [self.get_description(text)],
        ]

        for key, value in data.items():
            if not all(value.get(req) for req in required):
                continue

            observed_expected_corr = (
                self.plot_observed_vs_expected_correlations(
                    value, counts_df))
            detected_vs_undetected = (
                self.plot_detected_vs_undetected_references(
                    value, counts_df))
            observations_vs_expected_molarity = (
                self.plot_observations_vs_expected_molarity(
                    value, counts_df))

            plots.append([Div(text=f'<h3>{key}</h3>')])
            plots.append([observed_expected_corr, detected_vs_undetected])
            plots.append([observations_vs_expected_molarity])

        main = layout(plots, sizing_mode="scale_width")
        return Panel(child=main, title="Control")

    def plot_base_pairs(self, data):
        """Plot_base_pairs."""
        base_pairs = {
            self.abbreviate_name(k, data): v['base_pairs']
            for k, v in data.items()
        }

        data = pd.Series(base_pairs).reset_index(
            name='value').rename(columns={'index': 'group'})
        data['angle'] = data['value']/data['value'].sum() * 2 * math.pi
        if Category20c.get(len(base_pairs)):
            data['color'] = Category20c[len(base_pairs)]
        else:
            data['color'] = Colors.light_cornflower_blue
        data['percentage'] = (data['value']/data['value'].sum())
        data['megabases'] = (data['value'] / 1000000)

        plot = figure(
            plot_height=200,
            title="Base pairs breakdown",
            tools="hover",
            tooltips="@group: @value",
            x_range=(-0.5, 1.0),
            toolbar_location=None
        )

        plot.annular_wedge(
            x=0,
            y=1,
            inner_radius=0.1,
            outer_radius=0.2,
            start_angle=cumsum('angle', include_zero=True),
            end_angle=cumsum('angle'),
            line_color="white",
            fill_color='color',
            legend_field='group',
            source=data
        )

        plot.axis.axis_label = None
        plot.axis.visible = False
        plot.grid.grid_line_color = None
        plot.min_border_left = 0
        plot.legend.visible = True

        hover = plot.select({'type': HoverTool})
        hover.tooltips = [
            ('group', '@group'),
            ('percentage', '@percentage{0.00%}'),
            ('count (mb)', '@megabases{0.00}')
        ]

        total = "Total: {:.2f}mb".format(data['megabases'].sum())
        self.add_plot_title(plot, {}, total)
        self.style_plot(plot)
        return plot

    def plot_alignment_summary(self, data):
        """Plot_alignment_summary."""
        labels = ['Primary', 'Secondary', 'Supplementary']
        counts = [
            data['primary_count'],
            data['secondary_count'],
            data['supplementary_count']
        ]

        if not sum(counts):
            return

        plot = figure(
            x_range=labels,
            plot_height=350,
            output_backend="webgl",
            toolbar_location=None,
            x_axis_label='Outcome',
            y_axis_label='Count',
            tools=""
        )
        plot.vbar(
            x=labels, top=counts,
            width=0.9, fill_color=Colors.light_cornflower_blue)
        plot.xgrid.grid_line_color = None
        plot.y_range.start = 0

        total = f"Total observations: {data['observations']}"
        self.style_plot(plot)
        self.add_plot_title(plot, data, total)
        return plot

    def plot_accuracy_distribution(self, data):
        """Plot_accuracy_distribution."""
        counts = []
        for index, item in enumerate(data['alignment_accuracies']):
            for i in range(item):
                counts.append(index/10)

        plot = hist.histogram(
            [counts],
            bins=100,
            height=300,
            width=400,
            xlim=(0, 100),
            colors=[Colors.light_cornflower_blue],
            x_axis_label='Accuracy %',
            y_axis_label='Count'
        )

        median = f"Median: {data['median_accuracy']}%"
        self.style_plot(plot)
        self.add_plot_title(plot, data, median)
        return plot

    def plot_coverage_distribution(self, data):
        """Plot_coverage_distribution."""
        counts = []
        for index, item in enumerate(data['alignment_coverages']):
            for i in range(item):
                counts.append(index)

        plot = hist.histogram(
            [counts],
            bins=101,
            height=300,
            width=400,
            xlim=(0, 100),
            colors=[Colors.light_cornflower_blue],
            x_axis_label='Coverage %',
            y_axis_label='Count',
        )

        cov80_perc = round(data['cov80_percent'], 2)
        cov80 = (
            f"Cov80: {data['cov80_count']} ({cov80_perc}%)"
        )
        self.style_plot(plot)
        self.add_plot_title(plot, data, cov80)
        return plot

    def plot_qscore_distribution(self, data):
        """Plot_qscore_distribution."""
        counts = []
        for index, item in enumerate(data['aligned_qualities']):
            for i in range(item):
                counts.append(index/10)

        plot = hist.histogram(
            [counts],
            bins=600,
            height=300,
            width=400,
            xlim=(0, 30),
            colors=[Colors.light_cornflower_blue],
            x_axis_label='Avg Quality',
            y_axis_label='Count',
        )

        median = f"Median: {int(data['median_quality'])}"
        self.style_plot(plot)
        self.add_plot_title(plot, data, median)
        return plot

    def plot_read_length_distribution(self, data):
        """Plot_read_length_distribution."""
        counts = []
        max_length = 0
        for index, item in enumerate(data['read_lengths']):
            for i in range(item):
                counts.append(index * 50)
                length = index * 50
                if length > max_length:
                    max_length = length

        plot = hist.histogram(
            [counts],
            bins=1000,
            height=300,
            width=400,
            xlim=(0, max_length + 200),
            colors=[Colors.light_cornflower_blue],
            x_axis_label='Length',
            y_axis_label='Count',
        )

        n50 = f"Read n50: {int(data['read_n50'])}"
        self.style_plot(plot)
        self.add_plot_title(plot, data, n50)
        return plot

    def plot_observed_vs_expected_correlations(self, value, counts):
        """Plot_observed_vs_expected_correlations."""
        names = []
        obs = []
        exps = []

        countdict = counts.to_dict()['expected_count']
        for name, exp in countdict.items():
            names.append(name)
            exps.append(exp)
            obs.append(value['observed_references'].get(name, 0))

        data = pd.DataFrame(
            zip(names, obs, exps),
            columns=['Name', 'Observed', 'Expected']
        )

        data['log_obs'] = [math.log(y+1, 10) for y in data['Observed']]
        data['log_exp'] = [math.log(y+1, 10) for y in data['Expected']]

        model = LinearRegression().fit(
            np.array(data['log_exp'].values).reshape(-1, 1),
            data['log_obs']
        )

        regression_line = Slope(
            gradient=model.coef_[0],
            y_intercept=model.intercept_,
            line_color=Colors.light_cornflower_blue
        )

        plot = points.points(
            [data['log_exp'].tolist()],
            [data['log_obs'].tolist()],
            height=350,
            tools="",
            toolbar_location=None,
            output_backend="webgl",
            colors=[Colors.light_cornflower_blue],
            x_axis_label='log10(Expected Count)',
            y_axis_label='log10(Observed Count)',
            title='Expected vs Observed'
        )

        corrs = (
            "Spearmans: {:.2f}, (p = {:.2f})  "
            "Pearsons: {:.2f}, (p = {:.2f})".format(
                value['spearmans_rho'],
                value['spearmans_rho_pval'],
                value['pearson'],
                value['pearson_pval']
            ))
        plot.add_layout(regression_line)
        self.add_plot_title(plot, {}, corrs)
        self.style_plot(plot)
        return plot

    def plot_observations_vs_expected_molarity(self, value, counts):
        """Plot_observations_vs_expected_molarity."""
        names = []
        obs = []
        exps = []

        countdict = counts.to_dict()['expected_count']
        for name, exp in countdict.items():
            names.append(name)
            exps.append(exp)
            obs.append(value['observed_references'].get(name, 0))

        data = pd.DataFrame(
            zip(names, obs, exps),
            columns=['Name', 'Observed', 'Expected']
        )

        data['log_obs'] = [math.log(y+1, 10) for y in data['Observed']]
        data['log_exp'] = [math.log(y+1, 10) for y in data['Expected']]

        data.sort_values('Expected', inplace=True)

        plot = figure(
            x_range=data['Name'].tolist(),
            plot_height=200,
            toolbar_location=None,
            tools="",
            x_axis_label='Reference',
            y_axis_label='log10(Observed Count)',
            output_backend="webgl",
            title='Observations ordered by increasing expected count'
        )
        plot.vbar(
            x=data['Name'].tolist(),
            top=data['log_obs'].values, width=0.9,
            fill_color=Colors.light_cornflower_blue
        )
        plot.xgrid.grid_line_color = None
        plot.y_range.start = 0

        plot.xaxis.major_label_orientation = 3.14/2
        self.style_plot(plot)
        return plot

    def plot_detected_vs_undetected_references(self, value, counts_df):
        """plot_detected_vs_undetected_references."""
        names = ['Detected', 'Not Detected']
        detected = value['observed_reference_count']
        undetected = len(counts_df)
        data = [detected, undetected]

        plot = figure(
            x_range=names,
            toolbar_location=None,
            tools="",
            plot_height=350,
            x_axis_label='Outcome',
            y_axis_label='Count',
            output_backend="webgl",
            title='Detected vs Undetected references'
        )
        plot.vbar(
            x=names, top=data, width=0.9,
            fill_color=Colors.light_cornflower_blue
        )
        plot.xgrid.grid_line_color = None
        plot.y_range.start = 0

        detected = f"Detected: {detected} / {undetected}"
        self.add_plot_title(plot, {}, detected)
        self.style_plot(plot)
        return plot

    def get_description(self, desc):
        """Get Description."""
        styles = [
            "display:block;",
            "width:100%;",
            "padding:25px 0 0 0;",
            "font-size: 16px;",
            "margin-bottom: 0;"
        ]
        return Div(
            text=f'<p style="{"".join(styles)}" class="lead">{desc}</p>'
        )

    def style_plot(self, plot):
        """Style plots."""
        plot.margin = (10, 10, 40, 10)
        plot.background_fill_alpha = 0

    def abbreviate_name(self, name, data):
        """Abbreviate the name."""
        item = data[name]
        fasta = item['fasta']
        barcode = item['barcode']
        run_id = item['run_id']

        if len(fasta) > 20:
            fasta = fasta[0:20] + '...'

        if len(run_id) > 20:
            run_id = run_id[0:20] + '...'

        return f"{fasta} / {barcode} / {run_id}"

    def add_plot_title(self, plot, data, *extra_lines):
        """Annotate plot with titles."""
        for line in extra_lines:
            plot.add_layout(
                Title(
                    text=line,
                    text_line_height=0.5,
                    text_font_style='italic',
                    text_color='grey'
                ),
                'above',
            )
        for identkey in [
            'source', 'fasta', 'run_id',
                'barcode', 'reference', 'read_group']:
            if not data.get(identkey):
                continue
            plot.add_layout(
                Title(
                    text=data[identkey],
                    text_line_height=0.5
                ),
                'above'
            )


def main(args):
    """Entry point to create a mapping stats report."""
    report = HTMLReport(
        title='Alignment statistics',
        lead=(
            "Results generated through the wf-alignment "
            "nextflow workflow provided by Oxford Nanopore "
            "Technologies"))
    report.add_section(
        section=PlotMappingStats(
            json=args.json,
            counts=args.counts))
    report.write(args.name + '.html')


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Visualise mapula output',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)

    parser.add_argument(
        help="Input .json file to plot graphs from.",
        dest="json")

    parser.add_argument(
        '-c',
        help=(
            "Expected counts CSV. "
            "Required columns: reference,expected_count."),
        dest="counts",
        default=None,
        required=False)

    parser.add_argument(
        "-n",
        help="Prefix of the output files.",
        dest="name",
        required=False,
        default="report")
    return parser



"""Simple report components."""

import argparse
import json
import os

import pandas as pd

from aplanat.report import _maybe_new_report, HTMLReport
from aplanat.util import get_named_logger


_version_header = """
### Software versions

The table below highlights versions of key software used within the analysis.
"""


def version_table(
        versions, header=_version_header, report=None, th_color='#0084A9'):
    """Create a software version report from CSVs with information.

    :param versions: directory containing headerless CSVs with
        'software,version', or a single file.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    logger = get_named_logger('VerReport')
    report = _maybe_new_report(report)
    report.markdown(header)

    if os.path.isdir(versions):
        versions = [os.path.join(versions, x) for x in os.listdir(versions)]
    elif os.path.isfile(versions):
        versions = [versions]
    else:
        raise IOError('`versions` should be a file or directory.')

    verdata = list()
    for fname in versions:
        logger.info(f"Reading versions from file: {fname}.")
        try:
            with open(fname, 'r') as fh:
                for line in fh.readlines():
                    name, version = line.strip().split(',')
                    verdata.append((name, version))
        except Exception:
            logger.warning(f"Failed to read versions from: {fname}.")
    verdata = pd.DataFrame(verdata, columns=('Name', 'Version'))
    report.table(verdata, index=False, th_color=th_color)
    return report


_params_header = """
### Workflow parameters

The table below highlights values of the main parameters used in this analysis.
"""


def params_table(
        params, header=_params_header, report=None, th_color='#0084A9'):
    """Create a workflow parameter report from a JSON file.

    :param versions: Flat JSON file containing key, value pairs
        corresponding to workflow parameter 'key/name,value' pairs.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    report = _maybe_new_report(report)
    report.markdown(header)

    if not os.path.isfile(params):
        raise IOError('`params` should be a JSON file.')

    with open(params) as f:
        data = json.load(f)
        params = [(k, v) for k, v in data.items()]
        df_params = pd.DataFrame(params, columns=['Key', 'Value'])
        report.table(df_params, index=False, th_color=th_color)
        return report


def main(args):
    """Entry point to create demo report of functionality in this module."""
    report = HTMLReport(
        title="Simple component demo.",
        lead="A demonstration of small, simple reporting components.")
    report.add_section(section=version_table(args.versions))
    report.add_section(section=params_table(args.params))
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        "Simple report elements demo.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--versions",
        help=(
            "Headerless CSV containing 'software,version', "
            "or directory of such files."))
    parser.add_argument(
        "--params",
        help=(
            "A JSON file containing the workflow parameter "
            "key/values."))
    parser.add_argument(
        "--output", default="simple_components_report.html",
        help="Output HTML file.")
    return parser
